"""
Run IgDiscover

Calls Snakemake to produce all the output files.
"""
import sys
import logging
import resource
import platform
import pkg_resources
from snakemake import snakemake

from .config import Config
from . import CommandLineError
from ..utils import available_cpu_count
from .. import __version__


logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--dryrun', '-n', default=False, action='store_true',
        help='Do not execute anything')
    arg('--cores', '--jobs', '-j', metavar='N', type=int, default=available_cpu_count(),
        help='Run on at most N CPU cores in parallel. '
        'Default: Use as many cores as available (%(default)s)')
    arg('--keepgoing', '-k', default=False, action='store_true',
        help='If one job fails, finish the others.')
    arg('targets', nargs='*', default=[],
        help='File(s) to create. If omitted, the full pipeline is run.')


def main(args):
    run_snakemake(**vars(args))


def run_snakemake(
    dryrun=False,
    cores=available_cpu_count(),
    keepgoing=False,
    targets=None,
):
    try:
        config = Config.from_default_path()
    except FileNotFoundError as e:
        sys.exit("Pipeline configuration file {!r} not found. Please create it!".format(e.filename))

    print('IgDiscover version {} with Python {}. Configuration:'.format(__version__,
        platform.python_version()))
    for k, v in sorted(vars(config).items()):
        # TODO the following line is only necessary for non-YAML configurations
        if k.startswith('_'):
            continue
        print('   ', k, ': ', repr(v), sep='')
    sys.stdout.flush()

    old_root_handlers = logger.root.handlers
    root = logging.getLogger()
    root.handlers = []
    file_handler = logging.FileHandler("log.txt")
    root.addHandler(file_handler)

    snakefile_path = pkg_resources.resource_filename('igdiscover', 'Snakefile')
    success = snakemake(
        snakefile_path,
        snakemakepath='snakemake',  # Needed in snakemake 3.9.0
        dryrun=dryrun,
        cores=cores,
        keepgoing=keepgoing,
        printshellcmds=True,
        targets=targets,
    )
    logger.root.handlers = old_root_handlers

    if sys.platform == 'linux' and not dryrun:
        cputime = resource.getrusage(resource.RUSAGE_SELF).ru_utime
        cputime += resource.getrusage(resource.RUSAGE_CHILDREN).ru_utime
        h = int(cputime // 3600)
        m = (cputime - h * 3600) / 60
        print('Total CPU time: {}h {:.2f}m'.format(h, m))

    if not success:
        raise CommandLineError()



import sys
import os
import subprocess
from contextlib import contextmanager
from io import StringIO

import dnaio


@contextmanager
def capture_stdout():
    sio = StringIO()
    old_stdout = sys.stdout
    sys.stdout = sio
    yield sio
    sys.stdout = old_stdout


def datapath(path):
    return os.path.join(os.path.dirname(__file__), 'data', path)


def resultpath(path):
    return os.path.join(os.path.dirname(__file__), 'results', path)


def files_equal(path1, path2):
    return subprocess.call(['diff', '-u', path1, path2]) == 0


def convert_fastq_to_fasta(fastq, fasta):
    with dnaio.open(fastq) as inf:
        with dnaio.open(fasta, mode="w") as outf:
            for record in inf:
                outf.write(record)



import os
import sys
import contextlib
import shutil
from pathlib import Path

from xopen import xopen
import pytest

from igdiscover.__main__ import main
from .utils import datapath, resultpath, files_equal, convert_fastq_to_fasta
from igdiscover.cli.init import run_init
from igdiscover.cli.config import print_configuration, modify_configuration
from igdiscover.cli.run import run_snakemake
from igdiscover.cli.clonotypes import run_clonotypes


@pytest.fixture
def run(tmpdir):
    def _run(args, expected):
        """
        Run IgDiscover, redirecting stdout to a temporary file.
        Then compare the output with the contents of an expected file.
        """
        outpath = str(tmpdir.join('output'))
        print('Running:', ' '.join(args))
        with open(outpath, 'w') as f:
            old_stdout = sys.stdout
            sys.stdout = f
            main(args)
            sys.stdout = old_stdout
        assert files_equal(expected, outpath)

    return _run


@pytest.fixture
def pipeline_dir(tmp_path):
    """An initialized pipeline directory"""
    pipeline_path = tmp_path / "initializedpipeline"
    init_testdata(pipeline_path)
    return pipeline_path


def init_testdata(directory):
    run_init(
        database="testdata/database",
        reads1="testdata/reads.1.fastq.gz",
        directory=str(directory),
    )
    with chdir(directory):
        modify_configuration([("barcode_length_3prime", "21")])


@contextlib.contextmanager
def chdir(path):
    previous_path = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(previous_path)


@pytest.fixture(scope="session")
def filtered_tab_session(tmp_path_factory):
    """Generate iteration-01/filtered.tab.gz"""

    pipeline_dir = tmp_path_factory.mktemp("pipedir") / "pipedir"
    init_testdata(pipeline_dir)
    with chdir(pipeline_dir):
        run_snakemake(targets=["iteration-01/filtered.tab.gz"])
    return pipeline_dir


@pytest.fixture
def has_filtered_tab(filtered_tab_session, tmp_path):
    """
    Give a fresh copy of a pipeline dir in which iteration-01/filtered.tab.gz
    is guaranteed to exist
    """
    pipeline_dir = tmp_path / "has_filtered_tab"
    shutil.copytree(
        filtered_tab_session,
        pipeline_dir,
        symlinks=True,
        ignore=shutil.ignore_patterns((".snakemake")),
    )
    return pipeline_dir


def test_main():
    with pytest.raises(SystemExit) as exc:
        main(['--version'])
    assert exc.value.code == 0


def test_group_by_barcode_only(run):
    args = ['group', '-b', '4', datapath('ungrouped.fasta')]
    run(args,  resultpath('grouped-by-barcode-only.fasta'))


def test_group_by_pseudo_cdr3(run):
    args = ['group', '-b', '4', '--pseudo-cdr3=-5:-2', '--trim-g', datapath('ungrouped.fasta')]
    run(args,  resultpath('grouped.fasta'))


def test_group_by_pseudo_cdr3_barcode_at_end(run):
    args = ['group', '-b', '-4', '--pseudo-cdr3=1:3', datapath('ungrouped.fasta')]
    run(args, resultpath('grouped2.fasta'))


def test_clusterplot(tmpdir):
    main(['clusterplot', '-m', '10', datapath('clusterplot.tab.gz'), str(tmpdir)])
    assert tmpdir.join('IGHV1-1801.png').check()


def test_igblast(run):
    args = ['igblast', '--threads=1', datapath('database/'), datapath('igblast.fasta')]
    run(args, resultpath('assigned.tab'))


def test_run_init(pipeline_dir):
    assert pipeline_dir.is_dir()
    assert (pipeline_dir / "igdiscover.yaml").exists()


def test_print_configuration(pipeline_dir):
    print_configuration(path=pipeline_dir / "igdiscover.yaml")


def test_modify_configuration(pipeline_dir):
    modify_configuration(
        settings=[("d_coverage", "12"), ("j_discovery.allele_ratio", "0.37")],
        path=str(pipeline_dir / "igdiscover.yaml"),
    )
    import ruamel.yaml
    with open(pipeline_dir / "igdiscover.yaml") as f:
        config = ruamel.yaml.safe_load(f)
    assert config["d_coverage"] == 12
    assert config["j_discovery"]["allele_ratio"] == 0.37


def test_dryrun(pipeline_dir):
    with chdir(pipeline_dir):
        run_snakemake(dryrun=True)


def test_primers(pipeline_dir):
    # Test whether specifying primer sequences leads to a SyntaxError
    with chdir(pipeline_dir):
        modify_configuration(
            settings=[
                ("forward_primers", "['CGTGA']"),
                ("reverse_primers", "['TTCAC']"),
            ],
        )
        run_snakemake(dryrun=True)


# TODO slow
def test_only_forward_primer(pipeline_dir):
    # issue #107 (broken symlink)
    with chdir(pipeline_dir):
        modify_configuration(settings=[("forward_primers", "['CGTGA']")])
        # Create some dummy files so we don’t need to run irrelevant steps of the pipeline
        r = Path("reads")
        r.mkdir()
        with xopen(r / "2-merged.fastq.gz", "w") as f:
            pass
        s = Path("stats")
        s.mkdir()
        with open(s / "merging-successful", "w") as f:
            pass
        with open(s / "reads.json", "w") as f:
            f.write('{"total": 0}')
        with open(s / "trimmed.json", "w") as f:
            f.write('{"trimmed": 0}')
        run_snakemake(targets=["reads/sequences.fasta.gz"])


def test_flash(pipeline_dir):
    # Test using FLASH and parsing its log output
    with chdir(pipeline_dir):
        modify_configuration(settings=[("merge_program", "flash")])
        run_snakemake(targets=["stats/reads.json"])
        # Ensure FLASH was actually run
        assert (pipeline_dir / "reads/2-flash.log").exists()


def test_snakemake_assigned_tab(has_filtered_tab):
    assert (has_filtered_tab / "iteration-01/filtered.tab.gz").exists()
    assert not (has_filtered_tab / "iteration-01/new_V_germline.tab").exists()


def test_snakemake_exact_tab(has_filtered_tab):
    with chdir(has_filtered_tab):
        run_snakemake(targets=["iteration-01/exact.tab"])
    assert (has_filtered_tab / "iteration-01/exact.tab").exists()


def test_snakemake_final(has_filtered_tab):
    with chdir(has_filtered_tab):
        run_snakemake(targets=["nofinal"])
    assert (has_filtered_tab / "iteration-01/new_V_germline.tab").exists()
    assert not (has_filtered_tab / "final/assigned.tab.gz").exists()

    with chdir(has_filtered_tab):
        run_snakemake()
    assert (has_filtered_tab / "final/assigned.tab.gz").exists()


def test_clonotypes(has_filtered_tab):
    run_clonotypes(has_filtered_tab / "iteration-01/assigned.tab.gz", limit=5)


def test_fastq_input(has_filtered_tab, tmp_path):
    # Use merged reads from already-run pipeline as input for a new run
    single_reads = has_filtered_tab / "reads" / "2-merged.fastq.gz"
    directory = tmp_path / "singleend-fastq"
    run_init(
        database="testdata/database",
        single_reads=str(single_reads),
        directory=str(directory),
    )
    with chdir(directory):
        modify_configuration([("barcode_length_3prime", "21")])
        run_snakemake(targets=["stats/reads.json"])


def test_fasta_input(has_filtered_tab, tmp_path):
    fasta_path = tmp_path / "justfasta.fasta"
    convert_fastq_to_fasta(
        has_filtered_tab / "reads" / "2-merged.fastq.gz",
        fasta_path,
    )
    directory = tmp_path / "singleend-fasta"
    run_init(
        database="testdata/database",
        single_reads=str(fasta_path),
        directory=str(directory),
    )
    with chdir(directory):
        modify_configuration([("barcode_length_3prime", "21")])
        run_snakemake(targets=["stats/reads.json"])



"""
Change configuration file

Example:
    igdiscover config --set iterations 1

If no --set option is given, print the current configuration.
"""
import os
import logging
import ruamel.yaml

from igdiscover.config import Config

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--set', nargs=2, default=[], metavar=('KEY', 'VALUE'), action='append',
        help='Set KEY to VALUE. Use KEY.SUBKEY[.SUBSUBKEY...] for nested keys.'
            'Can be given multiple times.')
    arg('--file', default=Config.DEFAULT_PATH,
        help='Configuration file to modify. Default: igdiscover.yaml in current directory.')


def main(args):
    if args.set:
        modify_configuration(args.set, args.file)
    else:
        print_configuration(args.file)


def modify_configuration(
    settings,
    path=Config.DEFAULT_PATH,
):
    with open(path) as f:
        config = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)
    for k, v in settings:
        if not isinstance(k, str) or not isinstance(v, str):
            raise ValueError("key and value must both be strings")
        v = ruamel.yaml.safe_load(v)
        # config[k] = v
        item = config
        # allow nested keys
        keys = k.split('.')
        for i in keys[:-1]:
            item = item[i]
        item[keys[-1]] = v
    tmpfile = path + '.tmp'
    with open(tmpfile, 'w') as f:
        print(ruamel.yaml.dump(config, Dumper=ruamel.yaml.RoundTripDumper), end='', file=f)
    os.rename(tmpfile, path)


def print_configuration(path=Config.DEFAULT_PATH):
    with open(path) as f:
        config = ruamel.yaml.safe_load(f)
    print(ruamel.yaml.dump(config), end='')
